package org.jvnet.updatecenter2;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.project.MavenProject;
import org.apache.maven.project.MavenProjectHelper;
import org.apache.tools.ant.taskdefs.Expand;
import org.apache.tools.ant.taskdefs.Execute;
import org.apache.tools.ant.Project;

import java.io.File;
import java.io.ByteArrayOutputStream;
import java.io.UnsupportedEncodingException;
import java.io.IOException;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;

/**
 * Builds an IPS package.
 *
 * @goal package
 * @phase package
 * @author Kohsuke Kawaguchi
 */
public class PackageMojo extends AbstractMojo {
    /**
     * Package definition file that drives the packaging process.
     *
     * @parameter default-value="src/proto.py"
     * @required
     * @deprecated As of release 0.4, replaced by {@link #prototype}
     */
    @Deprecated protected File proto;
    /**
     * Package definition file that drives the packaging process.
     *
     * @parameter
     */
    protected File prototype;
    /**
     * Directory that extracts makepkgs.py to.
     *
     * If this directory already exists, then the script is assumed to be already in there.
     * This is useful for testing modified versions of <tt>makepkgs.py</tt>. This also
     * implements the caching semantics.
     *
     * @parameter default-value="${project.basedir}/target/"
     * @since 0.4
     */
    protected File makePkgsDir;

    /**
     * Local directory to deposit the generated package.
     *
     * @parameter default-value="target/ips-repository"
     * @deprecated As of release 0.4, replaced by {@link #prototype}
     */
    protected File localRepository;

    /**
     * The base directory to be passed to <tt>makepkgs.py</tt>.
     * This determines where the relative paths in your <tt>prototype.py</tt> are resolved.
     *
     * @parameter default-value="${project.basedir}"
     * @deprecated As of release 0.4, replaced by {@link #baseDir}
     */
    @Deprecated protected File baseDirectory;

    /**
     * The base directory to be passed to <tt>makepkgs.py</tt>.
     * This determines where the relative paths in your <tt>prototype.py</tt> are resolved.
     *
     * @parameter
     * @since 0.4
     */
    protected File baseDir;
    
    /**
     * Path to generated IPS package file.
     *
     * @parameter default-value="target/${project.artifactId}-${project.version}.ipstgz"
     * @deprecated As of release 0.4, replaced by {@link #destFile}
     */
    @Deprecated protected File packageFile;

    /**
     * Path to generated IPS package file. The generated file will be in .tgz or .zip
     * format depending the value entered.
     *
     * @parameter
     * @since 0.4
     */
    protected File destFile;
    
    /**
     * Repository URL to deposit the generated package.
     *
     * <p>
     * If this parameter is non-null, this takes precedence over
     * the local repository parameter.
     *
     * @parameter
     * @deprecated As of release 0.4, replaced by {@link #repourl}
     */
    @Deprecated protected String remoteRepository;
    
    /**
     * Repository URL to deposit the generated package, it can be either
     * a remote repository(http://remote-repository) or local repository
     * (file://C:\repository\pkg).
     *
     * <p>
     * If this parameter is non-null, this takes precedence over
     * the local repository parameter.
     *
     * @parameter
     * @since 0.4
     */
    protected String repourl;

    /**
     * Attaches the created package file as an artifact of the build
     * (which in turn causes it to be installed to your local repository
     * or to be deployed remotely to a Maven repository.)
     *
     * @parameter default-value="true"
     */
    protected boolean attach;

    /**
     * Path to python executable. By default, the mojo downloads minimal
     * Python specifically packaged for UC2, but if you'd like to
     * use your own version, specify it here.
     *
     * <p>
     * The typical use case of this is to specify it in the command-line,
     * like <tt>mvn -Dpython=python2.5 install</tt>
     *
     * @parameter expression="${python}"
     */
    protected String python;

    /**
     * This is injected by Maven and not meant to be set by POM.
     *
     * @parameter expression="${project}"
     * @required
     */
    protected MavenProject project;

    /**
     * The OS for which the packages are targeted. You can specify 
     * more than one os element. The prototype file allows you to specify
     * the ostarget value for each file/dir/dirtrees. If the OS specified
     * in the ostarget element of makepkgs matches the name specified in 
     * the prototype file, that particular file/dir/dirtees in the 
     * prototype will be included in the package.
     * 
     * @parameter
     * @since 0.4
     */
    protected String[] ostarget;
    
    
    /**
     * The package fmri (name@version) to use for manifest file.
     * 
     * @parameter
     * @since 0.4
     */
    protected String manifestFMRI;
    /**
     * The manifest file to create the package fmri, multiple manifest-fmri 
     * can be specified.
     * 
     * @parameter
     * @since 0.4
     */
    protected File[] manifestFile;
    /**
     * The base directory for the source files of a manifest.
     * 
     * @parameter
     * @since 0.4
     */
    protected File manifestBase;
    
    /**
     * @component
     */
    private MavenProjectHelper projectHelper;

    public void execute() throws MojoExecutionException, MojoFailureException {
        try {
            File jar = locateMakepkgs();
            unjar(jar,makePkgsDir);
            completePkg(makePkgsDir);
            executeMakePkgs();
        } catch (IOException e) {
            throw new MojoExecutionException("makepkgs failed",e);
        } 
    }

    private void executeMakePkgs() throws MojoExecutionException, IOException {
        boolean createdArchive = false;
        String archiveType = "tgz";

        if (repourl == null) {
            repourl = remoteRepository;
        }
        if (destFile == null) {
            destFile = packageFile;
        }
        if (baseDir ==  null) {
            baseDir = baseDirectory;
        }
        if (prototype==null) {
            prototype = proto;
        }
        List<String> commandLine = new ArrayList<String>();
        commandLine.add(getPythonBin());
        commandLine.add(new File(makePkgsDir,"pkg-toolkit/pkg/bin/makepkgs.py").getPath());
        if(repourl != null) {
            commandLine.add("-s");
            commandLine.add(repourl);
        } else {
            commandLine.add("-d");
            commandLine.add(localRepository.getAbsolutePath());
        }
        if(destFile != null) {
            if(destFile.getName().endsWith(".tgz")) {
                commandLine.add("-t");
                commandLine.add(destFile.getAbsolutePath());
            } else if (destFile.getName().endsWith(".zip")) {
                commandLine.add("-z");
                commandLine.add(destFile.getAbsolutePath());
                archiveType = "zip";
            } else {
                commandLine.add("-t");
                commandLine.add(destFile.getAbsolutePath());
            }
            createdArchive = true;
        }
        commandLine.add("-b");
        commandLine.add(baseDir.getAbsolutePath());
        if(ostarget != null) {
            for(String os : ostarget) {
                commandLine.add("-o");
                commandLine.add(os);
            }
        }
        commandLine.addAll(Arrays.asList(
                "-D", "groupId="+project.getGroupId(),
                "-D", "artifactId="+project.getArtifactId(),
                "-D", "version="+project.getVersion()));

        commandLine.add(prototype.getAbsolutePath());

        if (manifestBase != null) {
            commandLine.add("--manifest-base");
            commandLine.add(manifestBase.getAbsolutePath());
        }
        
        if (manifestFMRI != null) {
            commandLine.add("--manifest-fmri");
            commandLine.add(manifestFMRI);
        }
        
        if (manifestFile != null) {
            for (File mFile : manifestFile) {
                commandLine.add("--manifest-file");
                commandLine.add(mFile.toString());
            }
        }
        getLog().debug("Invoking "+commandLine);
        Execute exec = new Execute();
        exec.setCommandline(commandLine.toArray(new String[commandLine.size()]));
        exec.setWorkingDirectory(project.getBasedir());
        if(exec.execute()!=0)
            throw new MojoExecutionException("makepkgs failed");

        if(createdArchive && attach)
            projectHelper.attachArtifact(project, archiveType, null, destFile);
    }

    /**
     * Figures out the path to Python executable, while accomodating the difference of
     * path in various platforms.
     */
    private String getPythonBin() {
        if (python!=null)   return python;

        // if null, we need to figure out where the UC2 python bundle is.
        for (String candidate : new String[]{
                "target/pkg-toolkit/pkg/python2.4-minimal/python.exe",
                "target/pkg-toolkit/pkg/python2.4-minimal/bin/python"}) {
            File f = new File(project.getBasedir(),candidate);
            if (f.exists())     return f.getPath();
        }
        // everything failed.
        return "python";
    }

    /**
     * Unjars the jar to the specified target directory.
     */
    private void unjar(File jar, File destDir) {
        getLog().info("Destdir ="+destDir);
        destDir.mkdirs();

        getLog().info("Extracting IPS to "+destDir);        

        Expand expand = new Expand();
        expand.setProject(new Project());
        expand.setSrc(jar);
        expand.setDest(destDir);
        expand.execute();
    }
    
    private void completePkg(File destDir) throws MojoExecutionException {
        String[] command = new String[3];
        command[0] = "java";
        command[1] = "-jar";
        command[2] = destDir.getPath()+"/pkg-toolkit/pkg/lib/pkg-bootstrap.jar";
        
        Execute exec = new Execute();
        exec.setCommandline(command);
        exec.setWorkingDirectory(new File(destDir,"pkg-toolkit"));
        try {
            if(exec.execute()!=0)
                throw new MojoExecutionException("makepkgs failed");        
        } catch (IOException e) {
            throw new MojoExecutionException("IOError", e);
        }
    }

    /**
     * Locates the jar file that makepkgs scripts are in.
     */
    private File locateMakepkgs() throws MalformedURLException {
        String resURL = getClass().getClassLoader().getResource("pkg-toolkit").toString();
        resURL = resURL.substring(4, resURL.lastIndexOf('!')); // cut off jar: and the file name portion
        return new File(decode(new URL(resURL).getPath()));
    }

    /**
     * Decode '%HH'.
     */
    private static String decode(String s) {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        for( int i=0; i<s.length();i++ ) {
            char ch = s.charAt(i);
            if(ch=='%') {
                baos.write(hexToInt(s.charAt(i+1))*16 + hexToInt(s.charAt(i+2)));
                i+=2;
                continue;
            }
            baos.write(ch);
        }
        try {
            return new String(baos.toByteArray(),"UTF-8");
        } catch (UnsupportedEncodingException e) {
            throw new Error(e); // impossible
        }
    }

    private static int hexToInt(int ch) {
        return Character.getNumericValue(ch);
    }
}
